/*
 *  * Copyright (c) 2004 Xilinx, Inc.  All rights reserved.
 *
 * Xilinx, Inc.
 * XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS" AS A
 * COURTESY TO YOU.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION AS
 * ONE POSSIBLE   IMPLEMENTATION OF THIS FEATURE, APPLICATION OR
 * STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS IMPLEMENTATION
 * IS FREE FROM ANY CLAIMS OF INFRINGEMENT, AND YOU ARE RESPONSIBLE
 * FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE FOR YOUR IMPLEMENTATION
 * XILINX EXPRESSLY DISCLAIMS ANY WARRANTY WHATSOEVER WITH RESPECT TO
 * THE ADEQUACY OF THE IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO
 * ANY WARRANTIES OR REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE
 * FROM CLAIMS OF INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * chess game code by Sheng-Kuo Lu
 * sklu.iic93g@nctu.edu.tw
 */

// Located in: microblaze_0/include/xparameters.h
#include "xparameters.h"

#include "xutil.h"
#include "xintc_l.h"
#include "xuartlite_l.h"
#include "xgpio_l.h"

#define PIECE_OFFSET 0xA00

/* Address of a particular chess piece on the screen (rows are 20) */
#define BOARD(c,r) \
   (((unsigned char *)(XPAR_VGA_BASEADDR))[(c) + ((r) << 4) + ((r) << 2)])
#define CHESS(c,r) (BOARD((c+6),(r+4)))

/* Start of bitmap memory */
#define PIECE ((unsigned int *)XPAR_VGA_BASEADDR + PIECE_OFFSET)

#define EMPTY 0x00
#define KING 0x01
#define QUEEN 0x02
#define BISHOP 0x03
#define KNIGHT 0x04
#define ROOK 0x05
#define PAWN 0x06
#define SELECT_OFFSET 0x08
#define WINFLAG 0x0f
#define CURSOR 0x10
#define ABS(v) (((v) < 0) ? (0-(v)) : (v))
#define GET_TYPE(c) ((c) & 0x0f)
#define GET_PPL(c) (((c) & 0x20) >> 5)
#define GET_HINT(c) ((GET_TYPE(c)) > 7)

/* Must be a power of two */
#define SCANCODE_BUFFER_SIZE 16
unsigned char scancode_buffer[SCANCODE_BUFFER_SIZE]; // act as a circular queue
int scancode_buffer_front = 0;
int scancode_buffer_rear = 0;

int row = 0, col = 0; //cursor location
unsigned char ps = 0; //piece select flag
unsigned char player = 0; //player flag
unsigned char selected; //the selected piece register
int rsel = 0, csel = 0; //the selected piece location
unsigned char gameover = 0; //game over flag
int prison[2] = {0, 0}; //the captured prison
int lastmove[5] = {0, 0, 0, 0, 0};
int moveking[2] = {0 ,0};
int movelrook[2] = {0, 0};
int moverrook[2] = {0, 0};

unsigned char get_character() //queue delete operation
{
  unsigned char result;
  result = scancode_buffer[scancode_buffer_front];
  scancode_buffer_front = (scancode_buffer_front + 1) & (SCANCODE_BUFFER_SIZE - 1);
  return result;
}

int keyboard_interrupt_count = 0;
unsigned int scan_code = 0;
unsigned int keyboard_bit = 11;
unsigned int pbit=1, cbit=0, start=0;

/* PS/2 Keyboard interrupt service routine */
void ps2_int_handler(void *baseaddr_p)
{
  keyboard_interrupt_count++;

/* We're reading a scancode and the keyboard clock fell: sample the data */
  cbit = XGpio_mReadReg(XPAR_PS2IO_BASEADDR, XGPIO_DATA_OFFSET);
  if ((pbit == 1) && (cbit == 0)) start = 1;
  if (start == 1) {
    scan_code = (scan_code >> 1) | (cbit << 9);
    if (--keyboard_bit == 0) {
      /* keycode is ready */
      if (scancode_buffer_front != ((scancode_buffer_rear + 1) & (SCANCODE_BUFFER_SIZE - 1))) {
        /* buffer is not full; add the character */
        scancode_buffer[scancode_buffer_rear] = scan_code;
        scancode_buffer_rear = (scancode_buffer_rear + 1) & (SCANCODE_BUFFER_SIZE - 1);
      }
      keyboard_bit = 11;
      start = 0;
    }
  }
  pbit = cbit;

  /* Acknowledge the interrupt */
  XGpio_mWriteReg(XPAR_PS2IO_BASEADDR, XGPIO_ISR_OFFSET, 1);
}

void check_hint() //check if any place can move before piece select
{
  int i, j;
  for (i=0;i<8;i++) {
    for (j=0;j<8;j++) {
      if (GET_TYPE(CHESS(i, j)) >= SELECT_OFFSET) {
	ps = 1;
	return;
      }
    }
  }
}

void clear_hint() //clear all move hint after a move done
{
  unsigned char hint;
  int i, j;
  for (i=0;i<8;i++) {
    for (j=0;j<8;j++) {
      hint = CHESS(i, j);
      if (GET_TYPE(hint) >= SELECT_OFFSET) {
	CHESS(i, j) = hint - SELECT_OFFSET;
      }
    }
  }
}

int find_king() //find and move cursor to the king
{
  int i, j;
  unsigned char finder;
  for (i=0; i<8; i++) {
    for (j=0; j<8; j++) {
      finder = CHESS(j, i);
      if ((GET_TYPE(finder) == KING) && (GET_PPL(finder) == player)) {
        row = i;
        col = j;
        return 1;
      }
    }
  }
  return 0; //no king found means game over
}

void check_bishop(int step) //checking bishop-like moveing direction
{
  unsigned char move;
  int i;
  for (i=1; i<=step; i++) {
    if (((rsel-i) < 0) || ((csel-i) < 0)) break;
    move = CHESS(csel-i, rsel-i);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel-i, rsel-i) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel-i, rsel-i) = move + SELECT_OFFSET;
      break;
    }
  }
  for (i=1; i<=step; i++) {
    if (((rsel-i) < 0) || ((csel+i) > 7)) break;
    move = CHESS(csel+i, rsel-i);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel+i, rsel-i) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel+i, rsel-i) = move + SELECT_OFFSET;
      break;
    }
  }
  for (i=1; i<=step; i++) {
    if (((rsel+i) > 7) || ((csel-i) < 0)) break;
    move = CHESS(csel-i, rsel+i);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel-i, rsel+i) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel-i, rsel+i) = move + SELECT_OFFSET;
      break;
    }
  }
  for (i=1; i<=step; i++) {
    if (((rsel+i) > 7) || ((csel+i) > 7)) break;
    move = CHESS(csel+i, rsel+i);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel+i, rsel+i) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel+i, rsel+i) = move + SELECT_OFFSET;
      break;
    }
  }
}

void check_rook(int step) //checking rook-like moving direction
{
  unsigned char move;
  int i;
  for (i=1; i<=step; i++) {
    if ((rsel-i) < 0) break;
    move = CHESS(csel, rsel-i);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel, rsel-i) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel, rsel-i) = move + SELECT_OFFSET;
      break;
    }
  }
  for (i=1; i<=step; i++) {
    if ((csel-i) < 0) break;
    move = CHESS(csel-i, rsel);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel-i, rsel) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel-i, rsel) = move + SELECT_OFFSET;
      break;
    }
  }
  for (i=1; i<=step; i++) {
    if ((csel+i) > 7) break;
    move = CHESS(csel+i, rsel);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel+i, rsel) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel+i, rsel) = move + SELECT_OFFSET;
      break;
    }
  }
  for (i=1; i<=step; i++) {
    if ((rsel+i) > 7) break;
    move = CHESS(csel, rsel+i);
    if (GET_TYPE(move) == EMPTY) {
      CHESS(csel, rsel+i) = (player << 5) | SELECT_OFFSET;
    } else {
      if (GET_PPL(move) != player) CHESS(csel, rsel+i) = move + SELECT_OFFSET;
      break;
    }
  }
}

void piece_select() //all move checking when a piece selected
{
  unsigned char move;
  int i;

  selected = CHESS(col, row);
  if (GET_PPL(selected) == player) {
    rsel = row;
    csel = col;
    switch (GET_TYPE(selected)) {
    case KING:
      check_bishop(1);
      check_rook(1);
      if (moveking[player] != 1) {
        if (movelrook[player] != 2) {
          for (i=1; i<4; i++) {
            if ((GET_TYPE(CHESS(csel-i, rsel)) & 0x07) != EMPTY) break;
          }
          if (i == 4) CHESS(csel-3, rsel) = (player << 5) | SELECT_OFFSET;
        }
        if (moverrook[player] != 2) {
          for (i=1; i<3; i++) {
            if ((GET_TYPE(CHESS(csel+i, rsel)) & 0x07) != EMPTY) break;
          }
          if (i == 3) CHESS(csel+2, rsel) = (player << 5) | SELECT_OFFSET;
        }
      }
      break;
    case QUEEN:
      check_bishop(7);
      check_rook(7);
      break;
    case BISHOP:
      check_bishop(7);
      break;
    case KNIGHT:
      if (((csel-2) >= 0) && ((rsel-1) >= 0)) {
        move = CHESS(csel-2, rsel-1);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel-2, rsel-1) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel-2, rsel-1) = move + SELECT_OFFSET;
        }
      }
      if (((csel-2) >= 0) && ((rsel+1) <= 7)) {
        move = CHESS(csel-2, rsel+1);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel-2, rsel+1) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel-2, rsel+1) = move + SELECT_OFFSET;
        }
      }
      if (((csel+2) <= 7) && ((rsel-1) >= 0)) {
        move = CHESS(csel+2, rsel-1);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel+2, rsel-1) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel+2, rsel-1) = move + SELECT_OFFSET;
        }
      }
      if (((csel+2) <= 7) && ((rsel+1) <= 7)) {
        move = CHESS(csel+2, rsel+1);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel+2, rsel+1) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel+2, rsel+1) = move + SELECT_OFFSET;
        }
      }
      if (((csel-1) >= 0) && ((rsel-2) >= 0)) {
        move = CHESS(csel-1, rsel-2);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel-1, rsel-2) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel-1, rsel-2) = move + SELECT_OFFSET;
        }
      }
      if (((csel-1) >= 0) && ((rsel+2) <= 7)) {
        move = CHESS(csel-1, rsel+2);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel-1, rsel+2) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel-1, rsel+2) = move + SELECT_OFFSET;
        }
      }
      if (((csel+1) <= 7) && ((rsel-2) >= 0)) {
        move = CHESS(csel+1, rsel-2);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel+1, rsel-2) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel+1, rsel-2) = move + SELECT_OFFSET;
        }
      }
      if (((csel+1) <= 7) && ((rsel+2) <= 7)) {
        move = CHESS(csel+1, rsel+2);
        if (GET_TYPE(move) == EMPTY) {
          CHESS(csel+1, rsel+2) = (player << 5) | SELECT_OFFSET;
        } else if (GET_PPL(move) != player) {
          CHESS(csel+1, rsel+2) = move + SELECT_OFFSET;
        }
      }
      break;
    case ROOK:
      check_rook(7);
      if ((csel == 0) && (movelrook[player] != 2)) movelrook[player] = 1;
      if ((csel == 7) && (moverrook[player] != 2)) moverrook[player] = 1;
      break;
    case PAWN:
      switch (player) {
      case 0:
        if (GET_TYPE(CHESS(csel, rsel+1)) == EMPTY) {
          CHESS(csel, rsel+1) = (player << 5) | SELECT_OFFSET;
          if ((rsel == 1) && (GET_TYPE(CHESS(csel, rsel+2)) == EMPTY)) CHESS(csel, rsel+2) = (player << 5) | SELECT_OFFSET;
        }
        move = CHESS(csel+1, rsel+1);
        if ((GET_PPL(move) != player) && (GET_TYPE(move) != EMPTY)) CHESS(csel+1, rsel+1) = move + SELECT_OFFSET;
        move = CHESS(csel-1, rsel+1);
        if ((GET_PPL(move) != player) && (GET_TYPE(move) != EMPTY)) CHESS(csel-1, rsel+1) = move + SELECT_OFFSET;
        if ((lastmove[0] == PAWN) && ((lastmove[1] - lastmove[3]) == 2) && (lastmove[3] == rsel) &&
            (ABS(lastmove[4] - csel) == 1))
          CHESS(lastmove[4], lastmove[3]+1) = (player << 5) | SELECT_OFFSET;
        break;
      case 1:
        if (GET_TYPE(CHESS(csel, rsel-1)) == EMPTY) {
          CHESS(csel, rsel-1) = (player << 5) | SELECT_OFFSET;
          if ((rsel == 6) && (GET_TYPE(CHESS(csel, rsel-2)) == EMPTY)) CHESS(csel, rsel-2) = (player << 5) | SELECT_OFFSET;
        }
        move = CHESS(csel+1, rsel-1);
        if ((GET_PPL(move) != player) && (GET_TYPE(move) != EMPTY)) CHESS(csel+1, rsel-1) = move + SELECT_OFFSET;
        move = CHESS(csel-1, rsel-1);
        if ((GET_PPL(move) != player) && (GET_TYPE(move) != EMPTY)) CHESS(csel-1, rsel-1) = move + SELECT_OFFSET;
        if ((lastmove[0] == PAWN) && ((lastmove[3] - lastmove[1]) == 2) && (lastmove[3] == rsel) &&
            (ABS(lastmove[4] - csel) == 1))
          CHESS(lastmove[4], lastmove[3]-1) = (player << 5) | SELECT_OFFSET;
        break;
      default:
      }
      break;
    default:
    }
    check_hint();
  }
}

void piece_move() //operations when a piece move
{
  unsigned char check;
  check = CHESS(col, row);
  if (GET_HINT(check)) {
    if (GET_TYPE(check) != SELECT_OFFSET) { //capture occur
      switch (player) {
      case 0:
        BOARD(19-(prison[0]++), 0) = check - SELECT_OFFSET;
        break;
      case 1:
        BOARD(prison[1]++, 14) = check - SELECT_OFFSET;
        break;
      default:
      }
    }
    CHESS(csel, rsel) = EMPTY;
    if ((lastmove[0] == PAWN) && (GET_TYPE(selected)== PAWN) && (col == lastmove[4]) &&
        (ABS(lastmove[3] - lastmove[1]) == 2) && (ABS(row - lastmove[3]) == 1) && (col != csel)) {
      switch (player) {
      case 0:
        check = CHESS(col, row-1);
        BOARD(19-(prison[0]++), 0) = check;
        CHESS(col, row-1) = EMPTY;
        break;
      case 1:
        check = CHESS(col, row+1);
        BOARD(prison[1]++, 14) = check;
        CHESS(col, row+1) = EMPTY;
        break;
      default:
      }
    }
    if ((row == ((player + 7) & 7)) && (GET_TYPE(selected) == PAWN)) {
      CHESS(col, row) = selected - 4; //pawn promotion
    } else {
      CHESS(col, row) = selected;
    }
    if ((GET_TYPE(selected) == KING) && (ABS(col - csel) > 1)) {
      if (col == 1) {
        CHESS(2, row) = CHESS(0, row);
        CHESS(0, row) = EMPTY;
      } else {
        CHESS(5, row) = CHESS(7, row);
        CHESS(7, row) = EMPTY;
      }
    }
    lastmove[0] = GET_TYPE(selected); //piece type
    lastmove[1] = rsel;               //start row
    lastmove[2] = csel;               //start column
    lastmove[3] = row;                //end row
    lastmove[4] = col;                //end column
    if (GET_TYPE(selected) == KING) {
      moveking[player] = 1;
    } else if (GET_TYPE(selected) == ROOK) {
      if (movelrook[player] == 1) {
        movelrook[player] = 2;
      } else if (moverrook[player] == 1) {
        moverrook[player] = 2;
      }
    }
    player = player ^ 1;
    if (player == 0) {
      BOARD(0, 0) = 0x07;
      BOARD(19,14) = 0x00;
    } else {
      BOARD(0, 0) = 0x00;
      BOARD(19,14) = 0x27;
    }
    ps = 0;
    clear_hint();
    gameover = find_king() ^ 1;
  }
}

void game_reset() //initial the game system to restart
{
  int i, j;
  for (i=0; i<20; i++) {
    for (j=0; j<15; j++) {
      BOARD(i, j) = 0x00;
    }
  }
  CHESS(0, 0) = ROOK;
  CHESS(1, 0) = KNIGHT;
  CHESS(2, 0) = BISHOP;
  CHESS(3, 0) = QUEEN;
  CHESS(4, 0) = KING;
  CHESS(5, 0) = BISHOP;
  CHESS(6, 0) = KNIGHT;
  CHESS(7, 0) = ROOK;
  for (i=0; i<8; i++)
    CHESS(i, 1) = PAWN;
  CHESS(0, 7) = ROOK | 0x20;
  CHESS(1, 7) = KNIGHT | 0x20;
  CHESS(2, 7) = BISHOP | 0x20;
  CHESS(3, 7) = QUEEN | 0x20;
  CHESS(4, 7) = KING | 0x20;
  CHESS(5, 7) = BISHOP | 0x20;
  CHESS(6, 7) = KNIGHT | 0x20;
  CHESS(7, 7) = ROOK | 0x20;
  for (i=0; i<8; i++)
    CHESS(i, 6) = PAWN | 0x20;
  gameover = 0;
  prison[0] = 0;
  prison[1] = 0;
  moveking[0] = 0;
  moveking[1] = 0;
  movelrook[0] = 0;
  movelrook[1] = 0;
  moverrook[0] = 0;
  moverrook[1] = 0;
  if (player == 0) {
    BOARD(0, 0) = 0x07;
    BOARD(19,14) = 0x00;
  } else {
    BOARD(0, 0) = 0x00;
    BOARD(19,14) = 0x27;
  }
}

int main (void) {
   int i, j, extended=0;
   unsigned char uc;
   char buf;

   /* Enable MicroBlaze interrupts */
   microblaze_enable_interrupts();
   /* Register the keyboard interrupt handler */
   XIntc_RegisterHandler(XPAR_OPB_INTC_0_BASEADDR, XPAR_OPB_INTC_0_SYSTEM_PS2C_INTR,
      (XInterruptHandler)ps2_int_handler, (void *)0);
   /* Start the interrupt controller */
   XIntc_mMasterEnable(XPAR_OPB_INTC_0_BASEADDR);
   /* Enable timer and keyboard interrupt requests in the interrupt controller */
   XIntc_mEnableIntr(XPAR_OPB_INTC_0_BASEADDR, XPAR_SYSTEM_PS2C_MASK);

  for (;;) {
    if (gameover) {
      for (i=4; i<12; i++) {
        BOARD(5, i) = ((player ^ 1) << 5) | WINFLAG;
        BOARD(14, i) = ((player ^ 1) << 5) | WINFLAG;
      }
      for (j=5; j<15; j++) {
        BOARD(j, 3) = ((player ^ 1) << 5) | WINFLAG;
        BOARD(j, 12) = ((player ^ 1) << 5) | WINFLAG;
      }
    }
    if (scancode_buffer_front != scancode_buffer_rear) {
      uc = get_character();
      if (extended == 1) {
        switch (uc) {
        case 0x75: //press up
          row = (row - 1) & 7;
          break;
        case 0x72: //press down
          row = (row + 1) & 7;
          break;
        case 0x6b: //press left
          col = (col - 1) & 7;
          break;
        case 0x74: //press right
          col = (col + 1) & 7;
          break;
        case 0x5a: //press enter
          if (ps == 0) {
            if (gameover == 0) piece_select();
          } else {
            if (gameover == 0) piece_move();
          }
          break;
        default:
        }
      } else {
        switch (uc) {
          case 0x04: //press F3
            game_reset();
            break;
          case 0x5a: //press enter
            if (ps == 0) {
              if (gameover == 0) piece_select();
            } else {
              if (gameover == 0) piece_move();
            }
            break;
          case 0x76: //press esc
            ps = 0;
            if (gameover == 0) clear_hint();
            break;
          default:
          }
      }
      if (uc == 0xe0) {
        extended = 1;
      }
      if (uc == 0xf0) {
        scancode_buffer_front = scancode_buffer_rear;
        extended = 0;
      }
    }
    buf = CHESS(col, row);
    for (i = 0 ; i < 2 ; i++) {
      if (i == 0) {
        CHESS(col, row) = CURSOR | buf;
      } else {
        CHESS(col, row) = buf;
      }
      for (j = 0 ; j < 300000 ; j++ )  ; /* delay */
    }
  }
  return 0;
}

